@@ -1,6 +1,6 @@
 /* gda-data-proxy.c
  *
- * Copyright (C) 2005 - 2006 Vivien Malerba
+ * Copyright (C) 2005 - 2007 Vivien Malerba
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -31,6 +31,7 @@
 #include "gda-util.h"
 #include "gda-marshal.h"
 #include "gda-data-access-wrapper.h"
+#include "gda-enum-types.h"
 
 /* 
  * Main static functions 
@@ -55,11 +56,11 @@ static gint                 gda_data_proxy_get_n_rows      (GdaDataModel *model)
 static gint                 gda_data_proxy_get_n_columns   (GdaDataModel *model);
 static GdaColumn           *gda_data_proxy_describe_column (GdaDataModel *model, gint col);
 static const GValue        *gda_data_proxy_get_value_at    (GdaDataModel *model, gint col, gint row);
-static guint                gda_data_proxy_get_attributes_at (GdaDataModel *model, gint col, gint row);
+static GdaValueAttribute    gda_data_proxy_get_attributes_at (GdaDataModel *model, gint col, gint row);
 static GdaDataModelIter    *gda_data_proxy_create_iter     (GdaDataModel *model);
 
 
-static guint                gda_data_proxy_get_access_flags(GdaDataModel *model);
+static GdaDataModelAccessFlags gda_data_proxy_get_access_flags(GdaDataModel *model);
 
 static gboolean             gda_data_proxy_set_value_at    (GdaDataModel *model, gint col, gint row, 
 							    const GValue *value, GError **error);
@@ -89,10 +90,12 @@ enum
         ROW_DELETE_CHANGED,
 	SAMPLE_SIZE_CHANGED,
 	SAMPLE_CHANGED,
+	PRE_CHANGES_APPLIED,
+	POST_CHANGES_APPLIED,
         LAST_SIGNAL
 };
 
-static gint gda_data_proxy_signals[LAST_SIGNAL] = { 0, 0, 0 };
+static gint gda_data_proxy_signals[LAST_SIGNAL] = { 0, 0, 0, 0, 0 };
 
 
 /* properties */
@@ -105,13 +108,13 @@ enum
 
 /*
  * Structure to hold the status and all the modifications made to a single
- * row. It is not possible to have ((model_row < 0) && !modif_values)
+ * row. It is not possible to have ((model_row < 0) && !modify_values)
  */
 typedef struct
 {
 	gint           model_row;    /* row index in the GdaDataModel, -1 if new row */
 	gboolean       to_be_deleted;/* TRUE if row is to be deleted */
-	GSList        *modif_values; /* list of RowValue structures */
+	GSList        *modify_values; /* list of RowValue structures */
 	GValue       **orig_values;  /* array of the original GValues, indexed on the column numbers */
 	gint           orig_values_size;
 } RowModif;
@@ -125,7 +128,7 @@ typedef struct
 	RowModif      *row_modif;    /* RowModif in which this structure instance appears */
 	gint           model_column; /* column index in the GdaDataModel */
         GValue        *value;        /* can also be GDA_TYPE_LIST for multiple values; values are owned here */
-        GValue        *attributes;   /* holds a GUINT of GdaValueAttribute */
+        GValue        *attributes;   /* holds flags of GdaValueAttribute */
 } RowValue;
 #define ROW_VALUE(x) ((RowValue *)(x))
 
@@ -146,14 +149,14 @@ typedef struct
 struct _GdaDataProxyPrivate
 {
 	GdaDataModel      *model; /* Gda model which is proxied */
-	GValue           **columns_attrs; /* Each GValue holds a GUINT of GdaValueAttribute to proxy cols. attributes */
+	GValue           **columns_attrs; /* Each GValue holds a flag of GdaValueAttribute to proxy cols. attributes */
 
 	gint               model_nb_cols; /* = gda_data_model_get_n_columns (model) */
 	gboolean           notify_changes;
 
 	GSList            *all_modifs; /* list of RowModif structures, for memory management */
 	GSList            *new_rows;   /* list of RowModif, no data allocated in this list */
-	GHashTable        *modif_rows;  /* key = model_row number, value = RowModif, NOT for new rows */
+	GHashTable        *modify_rows;  /* key = model_row number, value = RowModif, NOT for new rows */
 
 	gboolean           ignore_proxied_changes;
 	gboolean           proxy_has_changed;
@@ -218,14 +221,14 @@ proxy_row_to_model_row (GdaDataProxy *proxy, gint row)
 }
 
 static RowModif *
-find_row_modif_for_proxy_row (GdaDataProxy *proxy, gint proxy_row)
+find_row_modify_for_proxy_row (GdaDataProxy *proxy, gint proxy_row)
 {
 	RowModif *rm = NULL;
 	gint model_row;
 
 	model_row = proxy_row_to_model_row (proxy, proxy_row);
 	if (model_row >= 0)
-		rm = g_hash_table_lookup (proxy->priv->modif_rows, GINT_TO_POINTER (model_row));
+		rm = g_hash_table_lookup (proxy->priv->modify_rows, GINT_TO_POINTER (model_row));
 	else {
 		gint offset = proxy->priv->current_nb_rows + 
 			(proxy->priv->add_null_entry ? 1 : 0);
@@ -251,6 +254,7 @@ find_proxy_row_for_row_modif (GdaDataProxy *proxy, RowModif *rm)
 	}
 }
 
+
 /*
  * Free a RowModif structure
  *
@@ -262,14 +266,14 @@ row_modifs_free (RowModif *rm)
 	GSList *list;
 	gint i;
 		
-	list = rm->modif_values;
+	list = rm->modify_values;
 	while (list) {
 		if (ROW_VALUE (list->data)->value)
 			gda_value_free (ROW_VALUE (list->data)->value);
 		g_free (list->data);
 		list = g_slist_next (list);
 	}
-	g_slist_free (rm->modif_values);
+	g_slist_free (rm->modify_values);
 	
 	if (rm->orig_values) {
 		for (i=0; i < rm->orig_values_size; i++) {
@@ -293,7 +297,7 @@ row_modifs_new (GdaDataProxy *proxy, gint proxy_row)
 	RowModif *rm;
 	
 #ifdef GDA_DEBUG
-	rm = find_row_modif_for_proxy_row (proxy, proxy_row);
+	rm = find_row_modify_for_proxy_row (proxy, proxy_row);
 	if (rm) 
 		g_warning ("%s(): RowModif already exists for that proxy_row", __FUNCTION__);
 #endif
@@ -358,6 +362,26 @@ gda_data_proxy_get_type (void)
 	return type;
 }
 
+static gboolean
+pre_changes_accumulator (GSignalInvocationHint *ihint,
+			 GValue *return_accu,
+			 const GValue *handler_return,
+			 gpointer data)
+{
+        gboolean thisvalue;
+
+        thisvalue = g_value_get_boolean (handler_return);
+        g_value_set_boolean (return_accu, thisvalue);
+
+        return thisvalue; /* stop signal if 'thisvalue' is FALSE */
+}
+
+static gboolean
+m_pre_changes_applied (GdaDataProxy *proxy, gint row, gint proxied_row)
+{
+        return TRUE; /* defaults allows changes */
+}
+
 static void
 gda_data_proxy_class_init (GdaDataProxyClass *class)
 {
@@ -387,10 +411,26 @@ gda_data_proxy_class_init (GdaDataProxyClass *class)
                               G_STRUCT_OFFSET (GdaDataProxyClass, sample_changed),
                               NULL, NULL,
 			      gda_marshal_VOID__INT_INT, G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_INT);
-	
+	gda_data_proxy_signals [PRE_CHANGES_APPLIED] =
+		g_signal_new ("pre_changes_applied",
+                              G_TYPE_FROM_CLASS (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GdaDataProxyClass, pre_changes_applied),
+                              pre_changes_accumulator, NULL,
+                              gda_marshal_BOOLEAN__INT_INT, G_TYPE_BOOLEAN, 2, G_TYPE_INT, G_TYPE_INT);
+	gda_data_proxy_signals [POST_CHANGES_APPLIED] =
+		g_signal_new ("post_changes_applied",
+                              G_TYPE_FROM_CLASS (object_class),
+                              G_SIGNAL_RUN_FIRST,
+                              G_STRUCT_OFFSET (GdaDataProxyClass, post_changes_applied),
+                              NULL, NULL,
+			      gda_marshal_VOID__INT_INT, G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_INT);
+
 	class->row_delete_changed = NULL;
 	class->sample_size_changed = NULL;
 	class->sample_changed = NULL;
+	class->pre_changes_applied = m_pre_changes_applied;
+	class->post_changes_applied = NULL;
 
 	/* virtual functions */
 #ifdef GDA_DEBUG
@@ -407,8 +447,7 @@ gda_data_proxy_class_init (GdaDataProxyClass *class)
 	g_object_class_install_property (object_class, PROP_MODEL,
 					 g_param_spec_object ("model", _("Data model"), NULL,
                                                                GDA_TYPE_DATA_MODEL,
-							       (G_PARAM_READABLE | G_PARAM_WRITABLE |
-								G_PARAM_CONSTRUCT_ONLY)));
+							       (G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT)));
 	g_object_class_install_property (object_class, PROP_ADD_NULL_ENTRY,
 					 g_param_spec_boolean ("prepend_null_entry", NULL, NULL, FALSE,
 							       (G_PARAM_READABLE | G_PARAM_WRITABLE)));
@@ -449,7 +488,7 @@ static void
 gda_data_proxy_init (GdaDataProxy *proxy)
 {
 	proxy->priv = g_new0 (GdaDataProxyPrivate, 1);
-	proxy->priv->modif_rows = g_hash_table_new (NULL, NULL);
+	proxy->priv->modify_rows = g_hash_table_new (NULL, NULL);
 	proxy->priv->notify_changes = TRUE;
 	proxy->priv->ignore_proxied_changes = FALSE;
 	proxy->priv->proxy_has_changed = FALSE;
@@ -469,13 +508,14 @@ static gboolean idle_add_model_rows (GdaDataProxy *proxy);
 
 
 static void proxied_model_data_changed_cb (GdaDataModel *model, GdaDataProxy *proxy);
+static void proxied_model_reset_cb (GdaDataModel *model, GdaDataProxy *proxy);
+
 
 /**
  * gda_data_proxy_new
- * @dict: a #GdaDict object
- * @type: the #GdaDictType requested
+ * @model: Data model to be proxied
  *
- * Creates a new proxy of type @type
+ * Creates a new proxy for @model
  *
  * Returns: a new #GdaDataProxy object
  */
@@ -483,15 +523,12 @@ GObject *
 gda_data_proxy_new (GdaDataModel *model)
 {
 	GObject *obj;
-	GdaDataProxy *proxy;
 
 	g_return_val_if_fail (model && GDA_IS_DATA_MODEL (model), NULL);
 
 	obj = g_object_new (GDA_TYPE_DATA_PROXY, "dict", gda_object_get_dict (GDA_OBJECT (model)), 
 			    "model", model, NULL);
 
-	proxy = GDA_DATA_PROXY (obj);
-
 	return obj;
 }
 
@@ -501,6 +538,49 @@ destroyed_object_cb (GdaObject *obj, GdaDataProxy *proxy)
 	gda_object_destroy (GDA_OBJECT (proxy));
 }
 
+static void
+clean_proxy (GdaDataProxy *proxy)
+{
+	if (proxy->priv->idle_add_event_source) {
+		g_idle_remove_by_data (proxy);
+		proxy->priv->idle_add_event_source = 0;
+	}
+	
+	if (proxy->priv->columns) {
+		gint i;
+		for (i = 0; i < 2 * proxy->priv->model_nb_cols; i++)
+			g_object_unref (G_OBJECT (proxy->priv->columns[i]));
+		g_free (proxy->priv->columns);
+		proxy->priv->columns = NULL;
+	}
+	
+	if (proxy->priv->model) {
+		g_signal_handlers_disconnect_by_func (G_OBJECT (proxy->priv->model),
+						      G_CALLBACK (proxied_model_data_changed_cb), proxy);
+		g_signal_handlers_disconnect_by_func (G_OBJECT (proxy->priv->model),
+						      G_CALLBACK (proxied_model_reset_cb), proxy);
+
+		g_signal_handlers_disconnect_by_func (G_OBJECT (proxy->priv->model),
+						      G_CALLBACK (destroyed_object_cb), proxy);
+		g_object_unref (proxy->priv->model);
+		proxy->priv->model = NULL;
+	}
+	
+	if (proxy->priv->columns_attrs) {
+		gint i;
+		for (i = 0; i < proxy->priv->model_nb_cols; i++)
+			gda_value_free ((GValue *)(proxy->priv->columns_attrs[i]));
+		g_free (proxy->priv->columns_attrs);
+		proxy->priv->columns_attrs = NULL;
+	}
+	
+	if (proxy->priv->modify_rows) {
+		gda_data_proxy_cancel_all_changes (proxy);
+		g_hash_table_destroy (proxy->priv->modify_rows);
+		proxy->priv->modify_rows = NULL;
+	}
+}
+
 static void
 gda_data_proxy_dispose (GObject *object)
 {
@@ -512,40 +592,7 @@ gda_data_proxy_dispose (GObject *object)
 	proxy = GDA_DATA_PROXY (object);
 	if (proxy->priv) {
 		gda_object_destroy_check (GDA_OBJECT (object));
-
-		if (proxy->priv->idle_add_event_source) {
-			g_idle_remove_by_data (proxy);
-			proxy->priv->idle_add_event_source = 0;
-		}
-
-		if (proxy->priv->columns) {
-			gint i;
-			for (i = 0; i < 2 * proxy->priv->model_nb_cols; i++)
-				g_object_unref (G_OBJECT (proxy->priv->columns[i]));
-			g_free (proxy->priv->columns);
-			proxy->priv->columns = NULL;
-		}
-
-		if (proxy->priv->model) {
-			g_signal_handlers_disconnect_by_func (G_OBJECT (proxy->priv->model),
-							      G_CALLBACK (destroyed_object_cb), proxy);
-			g_object_unref (proxy->priv->model);
-			proxy->priv->model = NULL;
-		}
-
-		if (proxy->priv->columns_attrs) {
-			gint i;
-			for (i = 0; i < proxy->priv->model_nb_cols; i++)
-				gda_value_free ((GValue *)(proxy->priv->columns_attrs[i]));
-			g_free (proxy->priv->columns_attrs);
-			proxy->priv->columns_attrs = NULL;
-		}
-		
-		if (proxy->priv->modif_rows) {
-			gda_data_proxy_cancel_all_changes (proxy);
-			g_hash_table_destroy (proxy->priv->modif_rows);
-			proxy->priv->modif_rows = NULL;
-		}
+		clean_proxy (proxy);
 	}
 
 	/* parent class */
@@ -603,28 +650,27 @@ gda_data_proxy_set_property (GObject *object,
 			g_object_ref (model);
 			gda_object_connect_destroy (GDA_OBJECT (model), G_CALLBACK (destroyed_object_cb), object);
 			
-			/* if (gda_data_model_get_status (model) & GDA_DATA_MODEL_NEEDS_INIT_REFRESH)  */
-			/* gda_data_model_refresh (model, NULL); */
-			
 			proxy->priv->model_nb_cols = gda_data_model_get_n_columns (model);
 			
 			/* column attributes */
 			proxy->priv->columns_attrs = g_new0 (GValue *, proxy->priv->model_nb_cols);
 			for (col = 0; col < proxy->priv->model_nb_cols; col++) {
 				GdaColumn *column;
-				guint flags = GDA_VALUE_ATTR_IS_UNCHANGED;
+				GdaValueAttribute flags = GDA_VALUE_ATTR_IS_UNCHANGED;
 				
 				column = gda_data_model_describe_column (model, col);
 				if (gda_column_get_allow_null (column))
 					flags |= GDA_VALUE_ATTR_CAN_BE_NULL;
 				if (gda_column_get_default_value (column))
 					flags |= GDA_VALUE_ATTR_CAN_BE_DEFAULT;
-				proxy->priv->columns_attrs[col] = g_value_init (g_new0 (GValue, 1), G_TYPE_UINT);
-				g_value_set_uint (proxy->priv->columns_attrs[col], flags);
+				proxy->priv->columns_attrs[col] = g_value_init (g_new0 (GValue, 1), GDA_TYPE_VALUE_ATTRIBUTE);
+				g_value_set_flags (proxy->priv->columns_attrs[col], flags);
 			}
 			
 			g_signal_connect (G_OBJECT (model), "changed",
 					  G_CALLBACK (proxied_model_data_changed_cb), proxy);
+			g_signal_connect (G_OBJECT (model), "reset",
+					  G_CALLBACK (proxied_model_reset_cb), proxy);
 			
 			adjust_displayed_chunck (proxy);
 			break;
@@ -673,6 +719,8 @@ static void
 proxied_model_data_changed_cb (GdaDataModel *model, GdaDataProxy *proxy)
 {
 	gint nb_new_rows = 0;
+	gint nb_cols;
+
 	if (proxy->priv->ignore_proxied_changes) {
 		proxy->priv->proxy_has_changed = TRUE;
 		return;
@@ -690,7 +738,7 @@ proxied_model_data_changed_cb (GdaDataModel *model, GdaDataProxy *proxy)
 		gint model_row = ROW_MODIF (proxy->priv->all_modifs->data)->model_row;
 		row_modifs_free (ROW_MODIF (proxy->priv->all_modifs->data));
 		if (model_row >= 0)
-			g_hash_table_remove (proxy->priv->modif_rows, GINT_TO_POINTER (model_row));
+			g_hash_table_remove (proxy->priv->modify_rows, GINT_TO_POINTER (model_row));
 		proxy->priv->all_modifs = g_slist_delete_link (proxy->priv->all_modifs, proxy->priv->all_modifs);
 	}
 	if (proxy->priv->new_rows) {/* new rows removing (no memory de-allocation here, though) */
@@ -699,19 +747,31 @@ proxied_model_data_changed_cb (GdaDataModel *model, GdaDataProxy *proxy)
 		proxy->priv->new_rows = NULL;
 	}
 
-	/* take into account the new dimensions of the proxied data model */
-	if (proxy->priv->model_nb_cols != gda_data_model_get_n_columns (model)) {
-		proxy->priv->model_nb_cols = gda_data_model_get_n_columns (model);
-		
-		/* FIXME: reset all proxy's attributes linked to proxy->priv->model_nb_cols */
-		TO_IMPLEMENT;
-
-		/* WILL break as implementation is missing */
-		g_assert_not_reached ();
+	nb_cols = gda_data_model_get_n_columns (model);
+	if (proxy->priv->model_nb_cols != nb_cols) {
+		/* take into account the new dimensions of the proxied data model => perform a re-init of the object */
+		proxied_model_reset_cb (model, proxy);
 	}
+	else {
+		proxy->priv->deferred_nb_rows += nb_new_rows;
+		adjust_displayed_chunck (proxy);
+	}
+}
 
-	proxy->priv->deferred_nb_rows += nb_new_rows;
-	adjust_displayed_chunck (proxy);
+/* 
+ * called when the proxied model emits a "reset" signal
+ */
+static void
+proxied_model_reset_cb (GdaDataModel *model, GdaDataProxy *proxy)
+{
+	gboolean add_null_entry = proxy->priv->add_null_entry;
+	
+	g_object_ref (G_OBJECT (model));
+	clean_proxy (proxy);
+	gda_data_proxy_init (proxy);
+	g_object_set (G_OBJECT (proxy), "model", model, "prepend_null_entry", add_null_entry, NULL);
+	g_object_unref (G_OBJECT (model));
+	gda_data_model_reset (GDA_DATA_MODEL (proxy));
 }
 
 /**
@@ -795,7 +855,7 @@ find_or_create_row_modif (GdaDataProxy *proxy, gint proxy_row, gint col, RowValu
 	gint model_row;
 
 	model_row = proxy_row_to_model_row (proxy, proxy_row);
-	rm = find_row_modif_for_proxy_row (proxy, proxy_row);
+	rm = find_row_modify_for_proxy_row (proxy, proxy_row);
 
 	if (!rm) {
 		/* create a new RowModif */
@@ -803,14 +863,14 @@ find_or_create_row_modif (GdaDataProxy *proxy, gint proxy_row, gint col, RowValu
 		rm = row_modifs_new (proxy, proxy_row);
 		rm->model_row = model_row;
 		if (model_row >= 0)
-			g_hash_table_insert (proxy->priv->modif_rows, GINT_TO_POINTER (model_row), rm);
+			g_hash_table_insert (proxy->priv->modify_rows, GINT_TO_POINTER (model_row), rm);
 		proxy->priv->all_modifs = g_slist_prepend (proxy->priv->all_modifs, rm);
 	}
 	else {
 		/* there are already some modifications to the row, try to catch the RowValue if available */
 		GSList *list;
 		
-		list = rm->modif_values;
+		list = rm->modify_values;
 		while (list && !rv) {
 			if (ROW_VALUE (list->data)->model_column == col)
 				rv = ROW_VALUE (list->data);
@@ -872,13 +932,13 @@ gda_data_proxy_get_values (GdaDataProxy *proxy, gint proxy_row, gint *cols_index
  *
  * Returns: the attribute
  */
-guint
+GdaValueAttribute
 gda_data_proxy_get_value_attributes (GdaDataProxy *proxy, gint proxy_row, gint col)
 {
 	gint model_row;
 	RowModif *rm;
 	gboolean value_has_modifs = FALSE;
-	guint flags = 0;
+	GdaValueAttribute flags = 0;
 	gint model_column;
 
 	g_return_val_if_fail (GDA_IS_DATA_PROXY (proxy), 0);
@@ -894,14 +954,14 @@ gda_data_proxy_get_value_attributes (GdaDataProxy *proxy, gint proxy_row, gint c
 			flags |= GDA_VALUE_ATTR_UNUSED;
 	}
 
-	rm = find_row_modif_for_proxy_row (proxy, proxy_row);
+	rm = find_row_modify_for_proxy_row (proxy, proxy_row);
 	if (rm) {
-		if (rm->modif_values) {
+		if (rm->modify_values) {
 			/* there are some modifications to the row */
 			GSList *list;
 			RowValue *rv = NULL;
 			
-			list = rm->modif_values;
+			list = rm->modify_values;
 			while (list && !rv) {
 				if (ROW_VALUE (list->data)->model_column == model_column)
 					rv = ROW_VALUE (list->data);
@@ -909,7 +969,7 @@ gda_data_proxy_get_value_attributes (GdaDataProxy *proxy, gint proxy_row, gint c
 			}
 			if (rv) {
 				value_has_modifs = TRUE;
-				flags |= g_value_get_uint (rv->attributes);
+				flags |= g_value_get_flags (rv->attributes);
 				if (rv->value && !gda_value_is_null (rv->value))
 					flags &= ~GDA_VALUE_ATTR_IS_NULL;
 				else
@@ -942,7 +1002,7 @@ gda_data_proxy_get_value_attributes (GdaDataProxy *proxy, gint proxy_row, gint c
  * flags (other flags are ignored).
  */
 void
-gda_data_proxy_alter_value_attributes (GdaDataProxy *proxy, gint proxy_row, gint col, guint alter_flags)
+gda_data_proxy_alter_value_attributes (GdaDataProxy *proxy, gint proxy_row, gint col, GdaValueAttribute alter_flags)
 {
 	gint model_col;
 
@@ -962,7 +1022,7 @@ gda_data_proxy_alter_value_attributes (GdaDataProxy *proxy, gint proxy_row, gint
 		g_assert (rm);
 		
 		if (alter_flags & GDA_VALUE_ATTR_IS_DEFAULT) {
-			guint flags;
+			GdaValueAttribute flags = 0;
 			if (!rv) {
 				/* create a new RowValue */
 				/*g_print ("New RV\n");*/
@@ -970,7 +1030,7 @@ gda_data_proxy_alter_value_attributes (GdaDataProxy *proxy, gint proxy_row, gint
 				rv->row_modif = rm;
 				rv->model_column = model_col;
 				rv->attributes = gda_value_copy (proxy->priv->columns_attrs [col]);
-				flags = g_value_get_uint (rv->attributes);
+				flags = g_value_get_flags (rv->attributes);
 							
 				rv->value = NULL;
 				flags &= ~GDA_VALUE_ATTR_IS_UNCHANGED;
@@ -979,17 +1039,17 @@ gda_data_proxy_alter_value_attributes (GdaDataProxy *proxy, gint proxy_row, gint
 				else
 					flags &= ~GDA_VALUE_ATTR_HAS_VALUE_ORIG;
 				
-				rm->modif_values = g_slist_prepend (rm->modif_values, rv);	
+				rm->modify_values = g_slist_prepend (rm->modify_values, rv);	
 			}
 			else {
-				flags = g_value_get_uint (rv->attributes);
+				flags = g_value_get_flags (rv->attributes);
 				if (rv->value) {
 					gda_value_free (rv->value);
 					rv->value = NULL;
 				}
 			}
 			flags |= GDA_VALUE_ATTR_IS_DEFAULT;
-			g_value_set_uint (rv->attributes, flags);
+			g_value_set_flags (rv->attributes, flags);
 
 			if (proxy->priv->notify_changes)
 				gda_data_model_row_updated ((GdaDataModel *) proxy, proxy_row);
@@ -1044,7 +1104,7 @@ gda_data_proxy_delete (GdaDataProxy *proxy, gint proxy_row)
 	g_return_if_fail (proxy_row >= 0);
 
 	model_row = proxy_row_to_model_row (proxy, proxy_row);
-	rm = find_row_modif_for_proxy_row (proxy, proxy_row);
+	rm = find_row_modify_for_proxy_row (proxy, proxy_row);
 	if (rm) {
 		if (! rm->to_be_deleted) {
 			if (rm->model_row < 0) {
@@ -1066,7 +1126,7 @@ gda_data_proxy_delete (GdaDataProxy *proxy, gint proxy_row)
 		/* the row is an existing row in the data model, create a new RowModif */
 		rm = row_modifs_new (proxy, proxy_row);
 		rm->model_row = model_row;
-		g_hash_table_insert (proxy->priv->modif_rows, GINT_TO_POINTER (model_row), rm);
+		g_hash_table_insert (proxy->priv->modify_rows, GINT_TO_POINTER (model_row), rm);
 		proxy->priv->all_modifs = g_slist_prepend (proxy->priv->all_modifs, rm);
 		rm->to_be_deleted = TRUE;
 		do_signal = TRUE;
@@ -1099,14 +1159,14 @@ gda_data_proxy_undelete (GdaDataProxy *proxy, gint proxy_row)
 	g_return_if_fail (proxy_row >= 0);
 
 	model_row = proxy_row_to_model_row (proxy, proxy_row);
-	rm = find_row_modif_for_proxy_row (proxy, proxy_row);
+	rm = find_row_modify_for_proxy_row (proxy, proxy_row);
 	if (rm) {
 		rm->to_be_deleted = FALSE;
-		if (!rm->modif_values) {
+		if (!rm->modify_values) {
 			/* get rid of that RowModif */
 			do_signal= TRUE;
 			
-			g_hash_table_remove (proxy->priv->modif_rows, GINT_TO_POINTER (model_row));
+			g_hash_table_remove (proxy->priv->modify_rows, GINT_TO_POINTER (model_row));
 			proxy->priv->all_modifs = g_slist_remove (proxy->priv->all_modifs, rm);
 			row_modifs_free (rm);
 		}
@@ -1140,7 +1200,7 @@ gda_data_proxy_row_is_deleted (GdaDataProxy *proxy, gint proxy_row)
 	g_return_val_if_fail (proxy->priv, FALSE);
 	g_return_val_if_fail (proxy_row >= 0, FALSE);
 
-	rm = find_row_modif_for_proxy_row (proxy, proxy_row);
+	rm = find_row_modify_for_proxy_row (proxy, proxy_row);
 	return rm && rm->to_be_deleted ? TRUE : FALSE;
 }
 
@@ -1163,7 +1223,7 @@ gda_data_proxy_row_is_inserted (GdaDataProxy *proxy, gint proxy_row)
 	g_return_val_if_fail (proxy->priv, FALSE);
 	g_return_val_if_fail (proxy_row >= 0, FALSE);
 
-	rm = find_row_modif_for_proxy_row (proxy, proxy_row);
+	rm = find_row_modify_for_proxy_row (proxy, proxy_row);
 	if (rm && (rm->model_row < 0))
 		return TRUE;
 	else
@@ -1287,16 +1347,16 @@ gda_data_proxy_append (GdaDataProxy *proxy)
 		GdaColumn *column;
 		const GValue *def;
 		RowValue *rv;
-		guint flags = 0;
+		GdaValueAttribute flags = 0;
 		
 		/* create a new RowValue */
 		/*g_print ("New RV\n");*/
 		rv = g_new0 (RowValue, 1);
 		rv->row_modif = rm;
 		rv->model_column = col;
-		rv->attributes = gda_value_new (G_TYPE_UINT);
+		rv->attributes = gda_value_new (GDA_TYPE_VALUE_ATTRIBUTE);
 		rv->value = NULL;
-		rm->modif_values = g_slist_prepend (rm->modif_values, rv);	
+		rm->modify_values = g_slist_prepend (rm->modify_values, rv);	
 		
 		column = gda_data_model_describe_column (proxy->priv->model, col);
 		def = gda_column_get_default_value (column);
@@ -1306,7 +1366,7 @@ gda_data_proxy_append (GdaDataProxy *proxy)
 				rv->value = gda_value_copy (def);
 		}
 		if (gda_column_get_allow_null (column)) {
-			guint attributes;
+			GdaValueAttribute attributes;
 			
 			attributes = gda_data_model_get_attributes_at (proxy->priv->model, col, -1);;
 			if (attributes & GDA_VALUE_ATTR_CAN_BE_NULL)
@@ -1316,7 +1376,7 @@ gda_data_proxy_append (GdaDataProxy *proxy)
 		if (gda_column_get_auto_increment (column))
 			flags |= (GDA_VALUE_ATTR_IS_DEFAULT | GDA_VALUE_ATTR_CAN_BE_DEFAULT);
 
-		g_value_set_uint (rv->attributes, flags);
+		g_value_set_flags (rv->attributes, flags);
 	}
 
 	/* signal row insertion */
@@ -1329,8 +1389,8 @@ gda_data_proxy_append (GdaDataProxy *proxy)
 /**
  * gda_data_proxy_cancel_row_changes
  * @proxy: a #GdaDataProxy object
+ * @proxy_row: the row to cancel changes
  * @col: the column to cancel changes, or less than 0 to cancel any change on the @row row
- * @row: the row to cancel changes
  *
  * Resets data at the corresponding row and column.
  */
@@ -1345,26 +1405,26 @@ gda_data_proxy_cancel_row_changes (GdaDataProxy *proxy, gint proxy_row, gint col
 	    (col < 0)) {
 		RowModif *rm;
 
-		rm = find_row_modif_for_proxy_row (proxy, proxy_row);
-		if (rm && rm->modif_values) {
+		rm = find_row_modify_for_proxy_row (proxy, proxy_row);
+		if (rm && rm->modify_values) {
 			/* there are some modifications to the row */
 			GSList *list;
 			RowValue *rv = NULL;
 			
-			list = rm->modif_values;
+			list = rm->modify_values;
 			while (list && (!rv || (col < 0))) {
 				if ((col < 0) || (ROW_VALUE (list->data)->model_column == col)) {
 					rv = ROW_VALUE (list->data);
 
 					/* remove this RowValue from the RowList */
-					rm->modif_values = g_slist_remove (rm->modif_values, rv);
-					if (!rm->modif_values && !rm->to_be_deleted) {
+					rm->modify_values = g_slist_remove (rm->modify_values, rv);
+					if (!rm->modify_values && !rm->to_be_deleted) {
 						/* remove this RowList as well */
 						proxy->priv->all_modifs = g_slist_remove (proxy->priv->all_modifs, rm);
 						if (rm->model_row < 0) 
 							proxy->priv->new_rows = g_slist_remove (proxy->priv->new_rows, rm);
 						else
-							g_hash_table_remove (proxy->priv->modif_rows, GINT_TO_POINTER (rm->model_row));
+							g_hash_table_remove (proxy->priv->modify_rows, GINT_TO_POINTER (rm->model_row));
 						row_modifs_free (rm);
 						rm = NULL;
 						
@@ -1376,7 +1436,7 @@ gda_data_proxy_cancel_row_changes (GdaDataProxy *proxy, gint proxy_row, gint col
 							gda_data_model_row_updated ((GdaDataModel *) proxy, proxy_row);
 					}
 					if (rm)
-						list = rm->modif_values;
+						list = rm->modify_values;
 					else
 						list = NULL;
 				}
@@ -1413,7 +1473,7 @@ gda_data_proxy_apply_row_changes (GdaDataProxy *proxy, gint proxy_row, GError **
 	gda_object_dump (proxy, 5);
 #endif
 
-	return commit_row_modif (proxy, find_row_modif_for_proxy_row (proxy, proxy_row), error);
+	return commit_row_modif (proxy, find_row_modify_for_proxy_row (proxy, proxy_row), error);
 }
 
 /*
@@ -1427,6 +1487,7 @@ commit_row_modif (GdaDataProxy *proxy, RowModif *rm, GError **error)
 	gboolean err = FALSE;
 	gboolean freedone = FALSE;
 	gboolean ignore_proxied_changes;
+	gboolean mod_ok;
 
 	if (!rm)
 		return TRUE;
@@ -1444,6 +1505,17 @@ commit_row_modif (GdaDataProxy *proxy, RowModif *rm, GError **error)
 	ignore_proxied_changes = proxy->priv->ignore_proxied_changes;
 	proxy->priv->ignore_proxied_changes = TRUE;
 
+	/* validate the changes to this row */
+        g_signal_emit (G_OBJECT (proxy),
+                       gda_data_proxy_signals[PRE_CHANGES_APPLIED],
+                       0, find_proxy_row_for_row_modif (proxy, rm), rm->model_row, &mod_ok);
+	if (!mod_ok) {
+		g_set_error (error, GDA_DATA_PROXY_ERROR, GDA_DATA_PROXY_COMMIT_CANCELLED,
+			     _("Modifications cancelled"));
+		return FALSE;
+	}
+
+	/* apply the changes */
 	if (rm->to_be_deleted) {
 		/* delete the row */
 		g_assert (rm->model_row >= 0);
@@ -1460,18 +1532,18 @@ commit_row_modif (GdaDataProxy *proxy, RowModif *rm, GError **error)
 			GValue *newvalue;
 			GValue **free_val;
 			
-			g_assert (rm->modif_values);
+			g_assert (rm->modify_values);
 			g_assert (rm->orig_values);
 			free_val = g_new0 (GValue *, proxy->priv->model_nb_cols);
 			for (i=0; i < rm->orig_values_size; i++) {
 				newvalue_found = FALSE;
 				newvalue = NULL;
 				
-				list = rm->modif_values;
+				list = rm->modify_values;
 				while (list && !newvalue_found) {
 					if (ROW_VALUE (list->data)->model_column == i) {
 						newvalue_found = TRUE;
-						if (g_value_get_uint (ROW_VALUE (list->data)->attributes) &
+						if (g_value_get_flags (ROW_VALUE (list->data)->attributes) &
 						    GDA_VALUE_ATTR_IS_DEFAULT)
 							newvalue = NULL;
 						else {
@@ -1506,15 +1578,15 @@ commit_row_modif (GdaDataProxy *proxy, RowModif *rm, GError **error)
 			GValue *newvalue;
 			GValue **free_val;
 			
-			g_assert (rm->modif_values);
+			g_assert (rm->modify_values);
 			free_val = g_new0 (GValue *, proxy->priv->model_nb_cols);
 			for (i = 0; i < proxy->priv->model_nb_cols; i++) {
 				newvalue = NULL;
 				
-				list = rm->modif_values;
+				list = rm->modify_values;
 				while (list && !newvalue) {
 					if (ROW_VALUE (list->data)->model_column == i) {
-						if (g_value_get_uint (ROW_VALUE (list->data)->attributes) &
+						if (g_value_get_flags (ROW_VALUE (list->data)->attributes) &
 						    GDA_VALUE_ATTR_IS_DEFAULT)
 							newvalue = NULL;
 						else {
@@ -1553,12 +1625,19 @@ commit_row_modif (GdaDataProxy *proxy, RowModif *rm, GError **error)
 			}
 		}
 	}
+
+	if (!err) {
+		/* signal row actually changed */
+		g_signal_emit (G_OBJECT (proxy),
+			       gda_data_proxy_signals[POST_CHANGES_APPLIED],
+			       0, find_proxy_row_for_row_modif (proxy, rm), rm->model_row);
 	
-	if (!err && !freedone) {
-		/* get rid of the commited change, where rm->model_row >= 0 */
-		proxy->priv->all_modifs = g_slist_remove (proxy->priv->all_modifs, rm);
-		g_hash_table_remove (proxy->priv->modif_rows, GINT_TO_POINTER (rm->model_row));
-		row_modifs_free (rm);
+		if (!freedone) {
+			/* get rid of the commited change, where rm->model_row >= 0 */
+			proxy->priv->all_modifs = g_slist_remove (proxy->priv->all_modifs, rm);
+			g_hash_table_remove (proxy->priv->modify_rows, GINT_TO_POINTER (rm->model_row));
+			row_modifs_free (rm);
+		}
 	}
 
 	proxy->priv->ignore_proxied_changes = ignore_proxied_changes;
@@ -1605,8 +1684,8 @@ gda_data_proxy_row_has_changed (GdaDataProxy *proxy, gint proxy_row)
 	g_return_val_if_fail (proxy->priv, FALSE);
 	g_return_val_if_fail (proxy_row >= 0, FALSE);
 
-	rm = find_row_modif_for_proxy_row (proxy, proxy_row);
-	return rm && (rm->modif_values || rm->to_be_deleted) ? TRUE : FALSE;
+	rm = find_row_modify_for_proxy_row (proxy, proxy_row);
+	return rm && (rm->modify_values || rm->to_be_deleted) ? TRUE : FALSE;
 }
 
 /**
@@ -1627,6 +1706,24 @@ gda_data_proxy_get_n_new_rows (GdaDataProxy *proxy)
         return g_slist_length (proxy->priv->new_rows);
 }
 
+/**
+ * gda_data_proxy_get_n_modified_rows
+ * @proxy: a #GdaDataProxy object
+ *
+ * Get the number of rows which have been modified in the proxy (the sum of rows existing in
+ * the proxied data model which have been modified, and new rows).
+ *
+ * Returns: the number of modified rows
+ */
+gint
+gda_data_proxy_get_n_modified_rows (GdaDataProxy *proxy)
+{
+	g_return_val_if_fail (GDA_IS_DATA_PROXY (proxy), 0);
+        g_return_val_if_fail (proxy->priv, 0);
+
+        return g_slist_length (proxy->priv->all_modifs);
+}
+
 #ifdef GDA_DEBUG
 static void
 gda_data_proxy_dump (GdaDataProxy *proxy, guint offset)
@@ -1986,6 +2083,8 @@ gda_data_proxy_apply_all_changes (GdaDataProxy *proxy, GError **error)
 	proxy->priv->ignore_proxied_changes = FALSE;
 
 	gda_data_model_send_hint (proxy->priv->model, GDA_DATA_MODEL_HINT_END_BATCH_UPDATE, NULL);
+	if (proxy->priv->proxy_has_changed) 
+		adjust_displayed_chunck (proxy);
 
 	return allok;
 }
@@ -2022,7 +2121,7 @@ gda_data_proxy_cancel_all_changes (GdaDataProxy *proxy)
 		
 		row_modifs_free (ROW_MODIF (proxy->priv->all_modifs->data));
 		if (model_row >= 0)
-			g_hash_table_remove (proxy->priv->modif_rows, GINT_TO_POINTER (model_row));
+			g_hash_table_remove (proxy->priv->modify_rows, GINT_TO_POINTER (model_row));
 		proxy->priv->all_modifs = g_slist_delete_link (proxy->priv->all_modifs, proxy->priv->all_modifs);
 		
 		if (model_row >= 0) {
@@ -2144,13 +2243,13 @@ gda_data_proxy_get_value_at (GdaDataModel *model, gint column, gint proxy_row)
 		gint model_col = column % proxy->priv->model_nb_cols;
 		gboolean value_has_modifs = FALSE;
 
-		rm = find_row_modif_for_proxy_row (proxy, proxy_row);
-		if (rm && rm->modif_values) {
+		rm = find_row_modify_for_proxy_row (proxy, proxy_row);
+		if (rm && rm->modify_values) {
 			/* there are some modifications to the row, see if there are some for the column */
 			GSList *list;
 			RowValue *rv = NULL;
 					
-			list = rm->modif_values;
+			list = rm->modify_values;
 			while (list && !rv) {
 				if (ROW_VALUE (list->data)->model_column == model_col)
 					rv = ROW_VALUE (list->data);
@@ -2182,7 +2281,7 @@ gda_data_proxy_get_value_at (GdaDataModel *model, gint column, gint proxy_row)
 		RowModif *rm;
 		gint model_col = column % proxy->priv->model_nb_cols;
 
-		rm = find_row_modif_for_proxy_row (proxy, proxy_row);
+		rm = find_row_modify_for_proxy_row (proxy, proxy_row);
 		if (rm) {
 			if (rm->orig_values)
 				retval = rm->orig_values [model_col];
@@ -2203,7 +2302,7 @@ gda_data_proxy_get_value_at (GdaDataModel *model, gint column, gint proxy_row)
 	return NULL;
 }
 
-static guint
+static GdaValueAttribute
 gda_data_proxy_get_attributes_at (GdaDataModel *model, gint col, gint row)
 {
 	g_return_val_if_fail (GDA_IS_DATA_PROXY (model), FALSE);
@@ -2250,7 +2349,7 @@ gda_data_proxy_create_iter (GdaDataModel *model)
 	return iter;
 }
 
-static guint
+static GdaDataModelAccessFlags
 gda_data_proxy_get_access_flags (GdaDataModel *model)
 {
 	GdaDataProxy *proxy;
@@ -2270,7 +2369,7 @@ gda_data_proxy_set_value_at (GdaDataModel *model, gint col, gint proxy_row, cons
 {
 	gint row;
 	GdaDataProxy *proxy;
-	guint att;
+	GdaValueAttribute att = 0;
 
 	g_return_val_if_fail (GDA_IS_DATA_PROXY (model), FALSE);
 	proxy = GDA_DATA_PROXY (model);
@@ -2317,13 +2416,13 @@ gda_data_proxy_set_value_at (GdaDataModel *model, gint col, gint proxy_row, cons
 			if (rm->orig_values && (col < rm->orig_values_size) && 
 			    ! gda_value_compare_ext ((GValue *) value, rm->orig_values [col])) {
 				/* remove the RowValue */
-				rm->modif_values = g_slist_remove (rm->modif_values, rv);
+				rm->modify_values = g_slist_remove (rm->modify_values, rv);
 				g_free (rv);
 				rv = NULL;
 			}
 			else {
 				/* simply alter the RowValue */
-				guint flags = g_value_get_uint (rv->attributes);
+				GdaValueAttribute flags = g_value_get_flags (rv->attributes);
 				
 				if (value && !gda_value_is_null ((GValue *) value)) {
 					flags &= ~GDA_VALUE_ATTR_IS_NULL;
@@ -2331,19 +2430,19 @@ gda_data_proxy_set_value_at (GdaDataModel *model, gint col, gint proxy_row, cons
 				}
 				else
 					flags |= GDA_VALUE_ATTR_IS_NULL;
-				g_value_set_uint (rv->attributes, flags);
+				g_value_set_flags (rv->attributes, flags);
 			}
 		}
 		else {
 			/* create a new RowValue */
 			/*g_print ("New RV\n");*/
-			guint flags;
+			GdaValueAttribute flags = 0;
 
 			rv = g_new0 (RowValue, 1);
 			rv->row_modif = rm;
 			rv->model_column = col;
 			rv->attributes = gda_value_copy (proxy->priv->columns_attrs [col]);
-			flags = g_value_get_uint (rv->attributes);
+			flags = g_value_get_flags (rv->attributes);
 
 			if (value && !gda_value_is_null ((GValue*) value)) {
 				rv->value = gda_value_copy ((GValue*) value);
@@ -2355,21 +2454,21 @@ gda_data_proxy_set_value_at (GdaDataModel *model, gint col, gint proxy_row, cons
 				flags |= GDA_VALUE_ATTR_HAS_VALUE_ORIG;
 			else
 				flags &= ~GDA_VALUE_ATTR_HAS_VALUE_ORIG;
-			g_value_set_uint (rv->attributes, flags);
-			rm->modif_values = g_slist_prepend (rm->modif_values, rv);
+			g_value_set_flags (rv->attributes, flags);
+			rm->modify_values = g_slist_prepend (rm->modify_values, rv);
 		}
 		
 		if (rv) {
-			guint flags = g_value_get_uint (rv->attributes);
+			GdaValueAttribute flags = g_value_get_flags (rv->attributes);
 			flags &= ~GDA_VALUE_ATTR_IS_UNCHANGED;
 			flags &= ~GDA_VALUE_ATTR_IS_DEFAULT;
-			g_value_set_uint (rv->attributes, flags);
+			g_value_set_flags (rv->attributes, flags);
 		}
 
-		if (!rm->to_be_deleted && !rm->modif_values && (rm->model_row >= 0)) {
+		if (!rm->to_be_deleted && !rm->modify_values && (rm->model_row >= 0)) {
 			/* remove that RowModif, it's useless */
 			/* g_print ("Removing RM\n"); */
-			g_hash_table_remove (proxy->priv->modif_rows, GINT_TO_POINTER (rm->model_row));
+			g_hash_table_remove (proxy->priv->modify_rows, GINT_TO_POINTER (rm->model_row));
 			proxy->priv->all_modifs = g_slist_remove (proxy->priv->all_modifs, rm);
 			row_modifs_free (rm);
 		}
